@@ -1,14 +1,12 @@
-BEGIN{$DB::single = 1; }
-
-package Cache::FastMmap::WithWin32;
+package Cache::FastMmap;
 
 =head1 NAME
 
-Cache::FastMmap::WithWin32 - Uses an mmap'ed file to act as a shared memory interprocess cache
+Cache::FastMmap - Uses an mmap'ed file to act as a shared memory interprocess cache
 
 =head1 SYNOPSIS
 
-  use Cache::FastMmap; # Yes, this is correct for this module
+  use Cache::FastMmap;
 
   # Uses vaguely sane defaults
   $Cache = Cache::FastMmap->new();
@@ -32,11 +30,6 @@ a basic LRU algorithm to keep the most used entries in the cache.
 
 =head1 DESCRIPTION
 
-This module is a fork of Cache::FastMmap by Rob Mueller to include the Win32
-port by Ash Berlin, cos i got fed up waiting for Rob to release it. 
-This distribution contains a file called Cache/FastMmap.pm, so installing this
-will be a drop in replacement for Cache::FastMmap
-
 In multi-process environments (eg mod_perl, forking daemons, etc),
 it's common to want to cache information, but have that cache
 shared between processes. Many solutions already exist, and may
@@ -171,8 +164,10 @@ comparisons to other modules.
 
 Cache::FastMmap uses mmap to map a file as the shared cache space,
 and fcntl to do page locking. This means it should work on most
-UNIX like operating systems, and now, thanks to Ash Berlin, will also work
-on Windows 2000 and above.
+UNIX like operating systems.
+
+Ash Berlin has written a Win32 layer using MapViewOfFile et al. to 
+provide support for Win32 platform.
 
 =head1 MEMORY SIZE
 
@@ -185,6 +180,69 @@ However, the OS will think your process is quite large, which might
 mean you hit some BSD::Resource or 'ulimits' you set previously that you
 thought were sane, but aren't anymore, so be aware.
 
+=head1 CACHE FILES AND OS ISSUES
+
+Because Cache::FastMmap uses an mmap'ed file, when you put values into
+the cache, you are actually "dirtying" pages in memory that belong to
+the cache file. Your OS will want to write those dirty pages back to
+the file on the actual physical disk, but the rate it does that at is
+very OS dependent.
+
+In Linux, you have some control over how the OS writes those pages
+back using a number of parameters in /proc/sys/vm
+
+  dirty_background_ratio
+  dirty_expire_centisecs
+  dirty_ratio
+  dirty_writeback_centisecs
+
+How you tune these depends heavily on your setup.
+
+As an interesting point, if you use a highmem linux kernel, a change
+between 2.6.16 and 2.6.20 made the kernel flush memory a LOT more.
+There's details in this kernel mailing list thread:
+L<http://www.uwsg.iu.edu/hypermail/linux/kernel/0711.3/0804.html>
+
+In most cases, people are not actually concerned about the persistence
+of data in the cache, and so are happy to disable writing of any cache
+data back to disk at all. Baically what they want is an in memory only
+shared cache. The best way to do that is to use a "tmpfs" filesystem
+and put all cache files on there.
+
+For instance, all our machines have a /tmpfs mount point that we
+create in /etc/fstab as:
+
+  none /tmpfs tmpfs defaults,noatime,size=1000M 0 0
+
+And we put all our cache files on there. The tmpfs filesystem is smart
+enough to only use memory as required by files actually on the tmpfs,
+so making it 1G in size doesn't actually use 1G of memory, it only uses
+as much as the cache files we put on it. In all cases, we ensure that
+we never run out of real memory, so the cache files effectively act 
+just as named access points to shared memory.
+
+Some people have suggested using anonymous mmaped memory. Unfortunately
+we need a file descriptor to do the fcntl locking on, so we'd have
+to create a separate file on a filesystem somewhere anyway. It seems
+easier to just create an explicit "tmpfs" filesystem.
+
+=head1 PAGE SIZE AND KEY/VALUE LIMITS
+
+To reduce lock contention, Cache::FastMmap breaks up the file
+into pages. When you get/set a value, it hashes the key to get a page,
+then locks that page, and uses a hash table within the page to
+get/store the actual key/value pair.
+
+One consequence of this is that you cannot store values larger than
+a page in the cache at all. Attempting to store values larger than
+a page size will fail (the set() function will return false).
+
+Also keep in mind that each page has it's own hash table, and that we
+store the key and value data of each item. So if you are expecting to
+store large values and/or keys in the cache, you should use page sizes
+that are definitely larger than your largest key + value size + a few
+kbytes for the overhead.
+
 =head1 USAGE
 
 Because the cache uses shared memory through an mmap'd file, you have
@@ -197,11 +255,13 @@ two main ways to do this:
 
 Create the cache in the parent process, and then when it forks, each
 child will inherit the same file descriptor, mmap'ed memory, etc and
-just work.
+just work. (BEWARE: This works just under UNIX as Win32 has no 
+concept of forking)
 
 =item *
 
-Explicitly connect up in each forked child to the share file
+Explicitly connect up in each forked child to the share file. (This is
+the only possible way under Win32)
 
 =back
 
@@ -227,16 +287,16 @@ use strict;
 use warnings;
 use bytes;
 
-our $VERSION = '1.16.1';
+our $VERSION = '1.28';
 
-use Cache::FastMmap::WithWin32::CImpl;
+use Cache::FastMmap::CImpl;
 
 use constant FC_ISDIRTY => 1;
 # }}}
 
 =item I<new(%Opts)>
 
-Create a new Cache::FastMmap::WithWin32 object.
+Create a new Cache::FastMmap object.
 
 Basic global parameters are:
 
@@ -244,9 +304,9 @@ Basic global parameters are:
 
 =item * B<share_file>
 
-File to mmap for sharing of data (default on unix: /tmp/sharefile-$pid-$time,
-default on windows depends on what windows thinks the temp dir is, but will 
-likely be: c:\Documents And Settings\User\Temp\sharefile-$pid-$time)
+File to mmap for sharing of data.
+default on unix: /tmp/sharefile-$pid-$time-$random
+default on windows: %TEMP%\sharefile-$pid-$time-$random
 
 =item * B<init_file>
 
@@ -264,6 +324,26 @@ the the cache file in an inconsistent state.
 Store values as raw binary data rather than using Storable to free/thaw
 data structures (default: 0)
 
+=item * B<compress>
+
+Compress the value (but not the key) before storing into the cache. If
+you set this to 1, the module will attempt to require the Compress::Zlib
+module and then use the memGzip() function on the value data before
+storing into the cache, and memGunzip() when retrieving data from the
+cache. Some initial testing shows that the uncompressing tends to be
+very fast, though the compressing can be quite slow, so it's probably
+best to use this option only if you know values in the cache are long
+lived and have a high hit rate. (default: 0)
+
+=item * B<enable_stats>
+
+Enable some basic statistics capturing. When enabled, every read to
+the cache is counted, and every read to the cache that finds a value
+in the cache is also counted. You can then retrieve these values
+via the get_statistics() call. This causes every read action to
+do a write on a page, which can cause some more IO, so it's
+disabled by default. (default: 0)
+
 =item * B<expire_time>
 
 Maximum time to hold values in the cache in seconds. A value of 0
@@ -328,7 +408,7 @@ Called as:
   $write_cb->($context, $Key, $Value, $ExpiryTime)
   
 In 'write_through' mode, it's always called as soon as a I<set(...)>
-is called on the Cache::FastMmap::WithWin32 class. In 'write_back' mode, it's
+is called on the Cache::FastMmap class. In 'write_back' mode, it's
 called when a value is expunged from the cache if it's been changed
 by a I<set(...)> rather than read from the underlying store with the
 I<read_cb> above.
@@ -347,7 +427,7 @@ Callback to delete data from the underlying data store.  Called as:
 
   $delete_cb->($context, $Key)
 
-Called as soon as I<remove(...)> is called on the Cache::FastMmap::WithWin32 class
+Called as soon as I<remove(...)> is called on the Cache::FastMmap class
 
 =item * B<cache_not_found>
 
@@ -402,30 +482,27 @@ exist, it defaults to 0.
 sub new {
   my $Proto = shift;
   my $Class = ref($Proto) || $Proto;
-  my %Args = @_;
+
+  # If first item is a hash ref, use it as arguments
+  my %Args = ref($_[0]) eq 'HASH' ? %{shift()} : @_;
 
   my $Self = {};
   bless ($Self, $Class);
 
-  # Initialise C cache code
-  my $Cache = Cache::FastMmap::WithWin32::CImpl::fc_new();
-
-  # We bless the returned scalar ref into the same namespace,
-  #  and store it in our own hash ref. We have to be sure
-  #  that we only call C functions on this scalar ref, and
-  #  only call PERL functions the hash ref we return
-  bless ($Cache, 'Cache::FastMmap::WithWin32::CImpl');
-
-
   # Work out cache file and whether to init
   my $share_file = $Args{share_file};
   if (!$share_file) {
-    $share_file = $Cache->fc_get_param('share_file');
+    my $tmp_dir = $ENV{TMPDIR} || "/tmp";
+    my $win_tmp_dir = $ENV{TEMP} || "c:\\";
+    $share_file = ($^O eq "MSWin32" ? "$win_tmp_dir\\sharefile" : "$tmp_dir/sharefile");
+    $share_file .= "-" . $$ . "-" . time . "-" . int(rand(100000));
   }
+  !ref($share_file) || die "share_file argument was a reference";
   $Self->{share_file} = $share_file;
 
-  my $init_file = $Args{init_file} || 0;
-  my $test_file = $Args{test_file} || 0;
+  my $init_file = $Args{init_file} ? 1 : 0;
+  my $test_file = $Args{test_file} ? 1 : 0;
+  my $enable_stats = $Args{enable_stats} ? 1 : 0;
 
   # Worth out unlink default if not specified
   if (!exists $Args{unlink_on_exit}) {
@@ -441,11 +518,17 @@ sub new {
       || die "Could not load Storable module: $@";
   }
 
+  # Compress stored values?
+  my $compress = $Self->{compress} = int($Args{compress} || 0);
+
+  # Need Compress::Zlib module if using compression
+  if ($compress) {
+    eval "use Compress::Zlib; 1;"
+      || die "Could not load Compress::Zlib module: $@";
+  }
+
   # Work out expiry time in seconds
-  my $expire_time = $Args{expire_time} || 0;
-  my %Times = (m => 60, h => 60*60, d => 24*60*60);
-  $expire_time *= $Times{$1} if $expire_time =~ s/([mhd])$//i;
-  $Self->{expire_time} = $expire_time = int($expire_time);
+  my $expire_time = $Self->{expire_time} = parse_expire_time($Args{expire_time});
 
   # Function rounds to the nearest power of 2
   sub RoundPow2 { return int(2 ** int(log($_[0])/log(2)) + 0.1); }
@@ -502,12 +585,21 @@ sub new {
     = @Args{qw(context read_cb write_cb delete_cb)};
   @$Self{qw(cache_not_found allow_recursive write_back)}
     = (@Args{qw(cache_not_found allow_recursive)}, $write_back);
-  @$Self{qw(empty_on_exit unlink_on_exit)}
-    = @Args{qw(empty_on_exit unlink_on_exit)};
+  @$Self{qw(empty_on_exit unlink_on_exit enable_stats)}
+    = (@Args{qw(empty_on_exit unlink_on_exit)}, $enable_stats);
 
   # Save pid
   $Self->{pid} = $$;
 
+  # Initialise C cache code
+  my $Cache = Cache::FastMmap::CImpl::fc_new();
+
+  # We bless the returned scalar ref into the same namespace,
+  #  and store it in our own hash ref. We have to be sure
+  #  that we only call C functions on this scalar ref, and
+  #  only call PERL functions the hash ref we return
+  bless ($Cache, 'Cache::FastMmap::CImpl');
+
   $Self->{Cache} = $Cache;
 
   # Setup cache parameters
@@ -518,6 +610,7 @@ sub new {
   $Cache->fc_set_param('expire_time', $expire_time);
   $Cache->fc_set_param('share_file', $share_file);
   $Cache->fc_set_param('start_slots', $start_slots);
+  $Cache->fc_set_param('enable_stats', $enable_stats);
 
   # And initialise it
   $Cache->fc_init();
@@ -570,13 +663,14 @@ sub get {
 
       # If not using raw values, use freeze() to turn data 
       $Val = freeze(\$Val) if !$Self->{raw_values};
+      $Val = Compress::Zlib::memGzip($Val) if $Self->{compress};
 
       # Get key/value len (we've got 'use bytes'), and do expunge check to
       #  create space if needed
       my $KVLen = length($_[1]) + (defined($Val) ? length($Val) : 0);
       $Self->_expunge_page(2, 1, $KVLen);
 
-      $Cache->fc_write($HashSlot, $_[1], $Val, 0);
+      $Cache->fc_write($HashSlot, $_[1], $Val, -1, 0);
     }
   }
 
@@ -585,9 +679,8 @@ sub get {
   $Cache->fc_unlock() unless $_[2] && $_[2]->{skip_unlock};
 
   # If not using raw values, use thaw() to turn data back into object
-  if (!$Self->{raw_values}) {
-    $Val = ${thaw($Val)} if defined $Val;
-  }
+  $Val = Compress::Zlib::memGunzip($Val) if defined($Val) && $Self->{compress};
+  $Val = ${thaw($Val)} if defined($Val) && !$Self->{raw_values};
 
   return $Val;
 }
@@ -600,16 +693,25 @@ I<%Options> is optional, and is used by get_and_set() to control
 the locking behaviour. For now, you should probably ignore it
 unless you read the code to understand how it works
 
+This method returns true if the value was stored in the cache,
+false otherwise. See the PAGE SIZE AND KEY/VALUE LIMITS section
+for more details.
+
 =cut
 sub set {
   my ($Self, $Cache) = ($_[0], $_[0]->{Cache});
 
   # If not using raw values, use freeze() to turn data 
   my $Val = $Self->{raw_values} ? $_[2] : freeze(\$_[2]);
+  $Val = Compress::Zlib::memGzip($Val) if $Self->{compress};
+
+  # Get opts, make compatiable with Cache::Cache interface
+  my $Opts = defined($_[3]) ? (ref($_[3]) ? $_[3] : { expire_time => $_[3] }) : undef;
+  my $expire_seconds = defined($Opts && $Opts->{expire_time}) ? parse_expire_time($Opts->{expire_time}) : -1;
 
   # Hash value, lock page
   my ($HashPage, $HashSlot) = $Cache->fc_hash($_[1]);
-  $Cache->fc_lock($HashPage) unless $_[3] && $_[3]->{skip_lock};
+  $Cache->fc_lock($HashPage) unless $Opts && $Opts->{skip_lock};
 
   # Are we doing writeback's? If so, need to mark as dirty in cache
   my $write_back = $Self->{write_back};
@@ -620,7 +722,7 @@ sub set {
   $Self->_expunge_page(2, 1, $KVLen);
 
   # Now store into cache
-  my $DidStore = $Cache->fc_write($HashSlot, $_[1], $Val, $write_back ? FC_ISDIRTY : 0);
+  my $DidStore = $Cache->fc_write($HashSlot, $_[1], $Val, $expire_seconds, $write_back ? FC_ISDIRTY : 0);
 
   # Unlock page
   $Cache->fc_unlock();
@@ -652,8 +754,13 @@ can just use:
 
   $Cache->get_and_set($Key, sub { return ++$_[1]; });
 
-The return value from this function is the new value stored back
-into the cache.
+In scalar context, the return value from this function is the *new* value
+stored back into the cache.
+
+In list context, a two item array is returned; the new value stored
+back into the cache and a boolean that's true if the value was stored
+in the cache, false otherwise. See the PAGE SIZE AND KEY/VALUE LIMITS
+section for more details.
 
 Notes:
 
@@ -667,7 +774,7 @@ operations lock the page and you may end up with a dead lock!
 =item *
 
 If your sub does a die/throws an exception, this will be caught
-to allow the pack to be unlocked, and then rethrown (1.15 onwards)
+to allow the page to be unlocked, and then rethrown (1.15 onwards)
 
 =back
 
@@ -678,23 +785,30 @@ sub get_and_set {
   my $Value = $Self->get($_[1], { skip_unlock => 1 });
   eval { $Value = $_[2]->($_[1], $Value); };
   my $Err = $@;
-  $Self->set($_[1], $Value, { skip_lock => 1 });
+  my $DidStore = $Self->set($_[1], $Value, { skip_lock => 1 });
   die $Err if $Err;
 
-  return $Value;
+  return wantarray ? ($Value, $DidStore) : $Value;
 }
 
-=item I<remove($Key)>
+=item I<remove($Key, [ \%Options ])>
 
 Delete the given key from the cache
 
+I<%Options> is optional, and is used by get_and_remove() to control
+the locking behaviour. For now, you should probably ignore it
+unless you read the code to understand how it works
+
 =cut
 sub remove {
   my ($Self, $Cache) = ($_[0], $_[0]->{Cache});
 
   # Hash value, lock page, read result
   my ($HashPage, $HashSlot) = $Cache->fc_hash($_[1]);
-  $Cache->fc_lock($HashPage);
+
+  # Lock is done only if we're not in the middle of a get_and_remove() operation.
+  $Cache->fc_lock($HashPage) unless $_[2] && $_[2]->{skip_lock};
+
   my ($DidDel, $Flags) = $Cache->fc_delete($HashSlot, $_[1]);
   $Cache->fc_unlock();
 
@@ -708,6 +822,23 @@ sub remove {
   return $DidDel;
 }
 
+=item I<get_and_remove($Key)>
+
+Atomically retrieve value of a Key while removing it from the cache.
+
+The page is locked while retrieving the $Key and is unlocked only after
+the value is removed, thus guaranteeing the value stored by someone else
+isn't removed by us.
+
+=cut
+sub get_and_remove {
+  my ($Self, $Cache) = ($_[0], $_[0]->{Cache});
+
+  my $Value = $Self->get($_[1], { skip_unlock => 1 });
+  my $DidDel = $Self->remove($_[1], { skip_lock => 1 });
+  return wantarray ? ($Value, $DidDel) : $Value;
+}
+
 =item I<clear()>
 
 Clear all items from the cache
@@ -770,14 +901,52 @@ sub get_keys {
 
   my $Mode = $_[1] || 0;
   return $Cache->fc_get_keys($Mode)
-    if $Mode <= 1 || ($Mode == 2 && $Self->{raw_values});
+    if $Mode <= 1 || ($Mode == 2 && $Self->{raw_values} && !$Self->{compress});
 
   # If we're getting values as well, and they're not raw, unfreeze them
   my @Details = $Cache->fc_get_keys(2);
-  for (@Details) { $_->{value} = ${thaw($_->{value})}; }
+  for (@Details) {
+    if (defined(my $Value = $_->{value})) {
+      $Value = Compress::Zlib::memGunzip($Value) if $Self->{compress};
+      $Value = ${thaw($Value)} if !$Self->{raw_values};
+      $_->{value} = $Value;
+    }
+  }
   return @Details;
 }
 
+=item I<get_statistics($Clear)>
+
+Returns a two value list of (nreads, nreadhits). This
+only works if you passed enable_stats in the constructor
+
+nreads is the total number of read attempts done on the
+cache since it was created
+
+nreadhits is the total number of read attempts done on
+the cache since it was created that found the key/value
+in the cache
+
+If $Clear is true, the values are reset immediately after
+they are retrieved
+
+=cut
+sub get_statistics {
+  my ($Self, $Cache) = ($_[0], $_[0]->{Cache});
+  my $Clear = $_[1];
+
+  my ($NReads, $NReadHits) = (0, 0);
+  for (0 .. $Self->{num_pages}-1) {
+    $Cache->fc_lock($_);
+    my ($PNReads, $PNReadHits) = $Cache->fc_get_page_details();
+    $NReads += $PNReads;
+    $NReadHits += $PNReadHits;
+    $Cache->fc_reset_page_details() if $Clear;
+    $Cache->fc_unlock();
+  }
+  return ($NReads, $NReadHits);
+}
+
 =item I<multi_get($PageKey, [ $Key1, $Key2, ... ])>
 
 The two multi_xxx routines act a bit differently to the
@@ -834,7 +1003,8 @@ sub multi_get {
     next unless $Found;
 
     # If not using raw values, use thaw() to turn data back into object
-    $Val = ${thaw($Val)} unless $Self->{raw_values};
+    $Val = Compress::Zlib::memGunzip($Val) if defined($Val) && $Self->{compress};
+    $Val = ${thaw($Val)} if defined($Val) && !$Self->{raw_values};
 
     # Save to return
     $KVs{$_} = $Val;
@@ -846,7 +1016,7 @@ sub multi_get {
   return \%KVs;
 }
 
-=item I<multi_set($PageKey, { $Key1 => $Value1, $Key2 => $Value2, ... })>
+=item I<multi_set($PageKey, { $Key1 => $Value1, $Key2 => $Value2, ... }, [ \%Options ])>
 
 Store specified key/value pair into cache
 
@@ -854,6 +1024,10 @@ Store specified key/value pair into cache
 sub multi_set {
   my ($Self, $Cache) = ($_[0], $_[0]->{Cache});
 
+  # Get opts, make compatiable with Cache::Cache interface
+  my $Opts = defined($_[3]) ? (ref($_[3]) ? $_[3] : { expire_time => $_[3] }) : undef;
+  my $expire_seconds = defined($Opts && $Opts->{expire_time}) ? parse_expire_time($Opts->{expire_time}) : -1;
+
   # Hash page key value, lock page
   my ($HashPage, $HashSlot) = $Cache->fc_hash($_[1]);
   $Cache->fc_lock($HashPage);
@@ -864,6 +1038,7 @@ sub multi_set {
 
     # If not using raw values, use freeze() to turn data 
     $Val = freeze(\$Val) unless $Self->{raw_values};
+    $Val = Compress::Zlib::memGzip($Val) if $Self->{compress};
 
     # Get key/value len (we've got 'use bytes'), and do expunge check to
     #  create space if needed
@@ -873,7 +1048,7 @@ sub multi_set {
 
     # Now hash key and store into page
     (undef, $HashSlot) = $Cache->fc_hash($FinalKey);
-    $Cache->fc_write($HashSlot, $FinalKey, $Val, 0);
+    my $DidStore = $Cache->fc_write($HashSlot, $FinalKey, $Val, $expire_seconds, 0);
   }
 
   # Unlock page
@@ -935,6 +1110,14 @@ sub _expunge_page {
   }
 }
 
+sub parse_expire_time {
+  my $expire_time = shift || '';
+  return 1 if $expire_time eq 'now';
+  return 0 if $expire_time eq 'never';
+  my %Times = ('' => 1, s => 1, m => 60, h => 60*60, d => 24*60*60, w => 7*24*60*60);
+  return $expire_time =~ /^(\d+)\s*([mhdws]?)/i ? $1 * $Times{$2} : 0;
+}
+
 sub DESTROY {
   my ($Self, $Cache) = ($_[0], $_[0]->{Cache});
 
@@ -954,15 +1137,9 @@ sub DESTROY {
 }
 
 sub CLONE {
-  die "Cache::FastMmap::WithWin32 does not support threads sorry";
+  die "Cache::FastMmap does not support threads sorry";
 }
 
-
-package #Hide From Pause
-   Cache::FastMmap;
-
-use base 'Cache::FastMmap::WithWin32';
-
 1;
 
 __END__
@@ -984,6 +1161,9 @@ This ensures that different runs/processes don't interfere with each other, but
 means you may not connect up to the file you expect. You should be choosing an
 explicit name in most cases.
 
+On Unix systems, you can pass in the environment variable TMPDIR to
+override the default directory of /tmp
+
 =item *
 
 The new option unlink_on_exit defaults to true if you pass a filename for the
@@ -994,6 +1174,15 @@ won't be.
 Otherwise the defaults seem sensible to cleanup unneeded share files rather than
 leaving them around to accumulate.
 
+=item After 1.28
+
+=over 4
+
+=item *
+
+Default share_file name is no longer /tmp/sharefile-$pid-$time 
+but /tmp/sharefile-$pid-$time-$random.
+
 =back
 
 =back
@@ -1009,15 +1198,11 @@ L<http://cpan.robm.fastmail.fm/cachefastmmap/>
 
 =head1 AUTHOR
 
-Original *nix version by Rob Mueller E<lt>L<mailto:cpan@robm.fastmail.fm>E<gt>
-
-Win32 port and refactoring by Ash Berlin L<< <ash@cpan.org> >>
+Rob Mueller L<mailto:cpan@robm.fastmail.fm>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2003-2007 by FastMail IP Partners
-
-Portions copyright Ash Berlin
+Copyright (C) 2003-2008 by FastMail IP Partners
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself. 
